/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"

#ifdef GRALECONFIG_SUPPORT_OPENCL

#include "renderserver.h"
#include "gravitationallens.h"
#include "openclkernel.h"
#include "constants.h"
#include <serut/memoryserializer.h>
#include <enut/packet.h>
#include <enut/tcpv4socket.h>
#include <enut/tcppacketsocket.h>
#ifdef GRALECONFIG_OPENCL_OPENCL_CL_H
	#include <OpenCL/cl.h>
#else
	#include <CL/cl.h>
#endif // GRALECONFIG_OPENCL_OPENCL_CL_H
#include <signal.h>
#include <time.h>
#include <iostream>
#include <string>

#include "debugnew.h"

using namespace grale;
bool endserverloop = false;

void SignalHandler(int val)
{
	endserverloop = true;
}

int main2(int argc, char *argv[])
{
	if (argc != 3)
	{
		std::cerr << "Usage: " << std::endl;
		std::cerr << std::string(argv[0]) << " serverIP portnumber" << std::endl; 
		return -1;
	}

	OpenCLKernel clKernel;

	if (!clKernel.init())
	{
		std::cerr << clKernel.getErrorString() << std::endl;
		return -1;
	}

	std::string mainCLProgram;

	mainCLProgram += "__kernel void lineRenderer(float2 startCoord, float xstep, int numX, float distFrac,\n";
	mainCLProgram += "                     __global const int *pIntParams,\n";
	mainCLProgram += "		     __global const float *pFloatParams,\n";
	mainCLProgram += "                     __global float *pResults)\n";
	mainCLProgram += "{\n";
	mainCLProgram += "	const int i = get_global_id(0);\n";
	mainCLProgram += "\n";
	mainCLProgram += "	if (i >= numX)\n";
	mainCLProgram += "		return;\n";
	mainCLProgram += "\n";
	mainCLProgram += "	float2 pos = { startCoord.x + xstep*((float)i), startCoord.y } ;\n";
	mainCLProgram += "\n";
	mainCLProgram += "	LensQuantities r = SUBROUTINENAME(pos, pIntParams, pFloatParams,-1,-1);\n";
	mainCLProgram += "\n";
	mainCLProgram += "	const int offset = i*5;\n";
	mainCLProgram += "\n";
	mainCLProgram += "	pResults[offset+0] = r.alphaX*distFrac;\n";
	mainCLProgram += "	pResults[offset+1] = r.alphaY*distFrac;\n";
	mainCLProgram += "	pResults[offset+2] = r.axx;\n";
	mainCLProgram += "	pResults[offset+3] = r.ayy;\n";
	mainCLProgram += "	pResults[offset+4] = r.axy;\n";
	mainCLProgram += "}\n";

	cl_mem clIntValues = 0;
	cl_mem clFloatValues = 0;
	cl_mem clResults = 0;
	float *pHostResults = 0;
	double *pDoubleResults = 0;
	cl_int err = 0;

	uint16_t port = (uint16_t)atoi(argv[2]);
	nut::IPv4Address address;
	nut::TCPv4Socket tcpSocket;
	nut::TCPPacketSocket connection(&tcpSocket, false, false, RENDERSERVER_MAXPACKSIZE, sizeof(uint32_t), RENDERSERVER_PACKID);

	address.setAddress(argv[1]);
	
	if (!tcpSocket.create())
	{
		std::cerr << "Error: " << std::endl;
		std::cerr << " Couldn't create TCP socket: " << tcpSocket.getErrorString() << std::endl;
		return -1;
	}

	std::cout << "Connecting to server" << std::endl;
	if (!tcpSocket.connect(address, port))
	{
		std::cerr << "Error: " << std::endl;
		std::cerr << " Couldn't open connection: " << connection.getErrorString() << std::endl;
		return -1;
	}

	// Ok, connection opened, send identification

	{
		uint8_t buf[sizeof(int32_t)];
		serut::MemorySerializer mser(0, 0, buf, sizeof(int32_t));

		mser.writeInt32(RENDERSERVER_COMMAND_RENDERER);

		std::cout << "Sending identification" << std::endl;
		if (!connection.write(buf, sizeof(int32_t)))
		{
			std::cerr << "Error: " << std::endl;
			std::cerr << " Error writing identification packet: " << connection.getErrorString() << std::endl;
			return -1;
		}

	}

	std::cout << "Installing signal handlers" << std::endl;

#ifndef WIN32
	signal(SIGTERM,SignalHandler);
	signal(SIGHUP,SignalHandler);
	signal(SIGQUIT,SignalHandler);
	signal(SIGABRT,SignalHandler);
	signal(SIGINT,SignalHandler);
#endif // WIN32

	GravitationalLens *lens = 0;
	Vector2D<double> bottomleft;
	Vector2D<double> topright;
	double D_s = 0;
	double D_ds = 0;
	double dx = 0;
	double dy = 0;
	int numx = 0;
	int numy = 0;
	uint8_t *linebuffer = 0;
	int linebuffersize = 0;
	double deflectionScale, potentialScale;
	
	// Ok, wait for incoming data from the server
	while (!endserverloop)
	{
		if (!connection.waitForData(1, 0))
		{
			endserverloop = true;
			continue;
		}

		if (connection.isDataAvailable())
		{
			size_t length = 0;

			connection.getAvailableDataLength(length);

			if (length == 0)
			{
				std::cerr << "Error: " << std::endl;
				std::cerr << " Server closed connection. " << std::endl;
				return -1;
			}
			else
			{	
				if (!connection.poll())
				{
					std::cerr << "Error: " << std::endl;
					std::cerr << " Error polling for new data: " << connection.getErrorString() << std::endl;
					return -1;
				}
				
				nut::Packet packet;

				while (connection.read(packet))
				{
					if (packet.getLength() < 4)
					{
						std::cerr << "Error: " << std::endl;
						std::cerr << " Received invalid packet from server" << std::endl;
						return -1;
					}
					
					serut::MemorySerializer mser(packet.getData(), packet.getLength(), 0, 0);
					int32_t cmd;
					
					mser.readInt32(&cmd);

					if (cmd == RENDERSERVER_COMMAND_ECHO)
					{
						// just ignore it
					}
					else if (cmd == RENDERSERVER_COMMAND_RENDER)
					{
						//std::cout << "Received render request" << std::endl;
				
						if (packet.getLength() != sizeof(int32_t)*2)
						{
							std::cerr << "Error: " << std::endl;
							std::cerr << " Received invalid render packet from server" << std::endl;
							return -1;
						}
						
						if (lens == 0)
						{
							std::cerr << "Error: " << std::endl;
							std::cerr << " Received render packet from server but no lens information has been received yet" << std::endl;
							return -1;
						}
						
						int32_t line;

						mser.readInt32(&line);
						
						if (line < 0 || line >= numy)
						{
							std::cerr << "Error: " << std::endl;
							std::cerr << " Received line number is out of range" << std::endl;
							return -1;
						}

						double x = bottomleft.getX();
						double y = bottomleft.getY()+(((double)line)*dy);
					
						serut::MemorySerializer mser2(0,0,linebuffer,linebuffersize);
						mser2.writeInt32(RENDERSERVER_COMMAND_RESULT);

						std::cout << "Rendering requested line using OpenCL" << line << std::endl;

						cl_int numX = numx;
						cl_float2 startPos = { (float)(x/deflectionScale), (float)(y/deflectionScale) };
						cl_float xstep = (float)(dx/deflectionScale);
						cl_float distFrac = (float)(D_ds/D_s);
						cl_kernel kernel = clKernel.getKernel();
						cl_command_queue command_queue = clKernel.getCommandQueue();

						err = clSetKernelArg(kernel, 0, sizeof(cl_float2), (void *)&startPos);
						err = clSetKernelArg(kernel, 1, sizeof(cl_float), (void *)&xstep);
						err = clSetKernelArg(kernel, 2, sizeof(cl_int), (void *)&numX);
						err = clSetKernelArg(kernel, 3, sizeof(cl_float), (void *)&distFrac);
						err = clSetKernelArg(kernel, 4, sizeof(cl_mem), (void *)&clIntValues);
						err = clSetKernelArg(kernel, 5, sizeof(cl_mem), (void *)&clFloatValues);
						err = clSetKernelArg(kernel, 6, sizeof(cl_mem), (void *)&clResults);

						size_t workSize[1] = { (size_t)numX };

						err = clEnqueueNDRangeKernel(command_queue, kernel, 1, NULL, workSize, NULL, 0, NULL, NULL);
						clFinish(command_queue);

						err = clEnqueueReadBuffer(command_queue, clResults, true, 0, sizeof(float)*(numX*5), pHostResults, 0, NULL, NULL);
						clFinish(command_queue);

						int numFloats = numx*5;

						for (int i = 0 ; i < numFloats ; i += 5)
						{
							pDoubleResults[i+0] = deflectionScale*pHostResults[i+0];
							pDoubleResults[i+1] = deflectionScale*pHostResults[i+1];
							pDoubleResults[i+2] = pHostResults[i+2];
							pDoubleResults[i+3] = pHostResults[i+3];
							pDoubleResults[i+4] = pHostResults[i+4];

						//	std::cerr << ((1.0-distFrac*pDoubleResults[i+2])*(1.0-distFrac*pDoubleResults[i+3])-distFrac*distFrac*pDoubleResults[i+4]*pDoubleResults[i+4]) << "\t";
						}
						std::cerr << std::endl;

						mser2.writeDoubles(pDoubleResults, numFloats);

						// Calculation ended, write result

						//std::cout << "Sending rendered line to server" << std::endl;
						if (!connection.write(linebuffer, linebuffersize))
						{
							std::cerr << "Error: " << std::endl;
							std::cerr << " Error writing render result: " << connection.getErrorString() << std::endl; 
							return -1;
						}
					}
					else if (cmd == RENDERSERVER_COMMAND_LENS)
					{
						//std::cout << "Received lens information" << std::endl;

						bool error = false;
						if (!mser.readInt32(&numx))
							error = true;
						if (!mser.readInt32(&numy))
							error = true;
						if (numx < 2 || numx > 100000)
							error = true;
						if (numy < 2 || numy > 100000)
							error = true;
						if (!mser.readDoubles(bottomleft.getComponents(), 2))
							error = true;
						if (!mser.readDoubles(topright.getComponents(), 2))
							error = true;
						if (!mser.readDouble(&D_s))
							error = true;
						if (!mser.readDouble(&D_ds))
							error = true;

						GravitationalLens *l = 0;
						std::string errstr;
				
						if (!GravitationalLens::read(mser,&l,errstr))
							error = true;
				
						if (mser.getBytesRead() != packet.getLength())
						{
							delete l;
							error = true;
						}
						
						if (error)
						{
							std::cerr << "Error: " << std::endl;
							std::cerr << " Received invalid lens information from server" << std::endl;
							return -1;
						}

						// No error, store new lens

						if (lens)
						{
							delete lens;
							delete [] linebuffer;
						}
						
						lens = l;
						linebuffersize = sizeof(int32_t)+sizeof(double)*5*numx;
						linebuffer = new uint8_t[linebuffersize];

						dx = (topright.getX()-bottomleft.getX())/((double)(numx-1));
						dy = (topright.getY()-bottomleft.getY())/((double)(numy-1));

						int numIntParams = 0;
						int numFloatParams = 0;

						if (!lens->getCLParameterCounts(&numIntParams, &numFloatParams))
						{
							std::cerr << lens->getErrorString() << std::endl;
							return -1;
						}

						int *pIntParams = new int[numIntParams+1]; // to avoid 0
						float *pFloatParams = new float[numFloatParams+1]; // idem
						
						deflectionScale = 1.0;
						potentialScale = 1.0;

						if (!lens->getSuggestedScales(&deflectionScale, &potentialScale))
						{
							deflectionScale = ANGLE_ARCSEC;
							potentialScale = ANGLE_ARCSEC*ANGLE_ARCSEC;
						}

						lens->getCLParameters(deflectionScale, potentialScale, pIntParams, pFloatParams);

						std::string programString;
						std::string routineName;
						std::string buildLog;
	
						programString += lens->getCLLensProgram(routineName);
						programString += mainCLProgram;	

						std::string::size_type pos = programString.find("SUBROUTINENAME");
						programString = programString.replace(pos, 14, routineName);

						std::cout << programString << std::endl;

						if (!clKernel.loadKernel(programString, "lineRenderer", buildLog))
						{
							std::cerr << clKernel.getErrorString() << std::endl;
							if (buildLog.length() > 0)
							{
								std::cerr << "Details: " << std::endl;
								std::cerr << buildLog << std::endl;
								std::cerr << std::endl;
							}
							return -1;
						}

						if (clIntValues)
							clReleaseMemObject(clIntValues);
						if (clFloatValues)
							clReleaseMemObject(clFloatValues);

						clIntValues = clCreateBuffer(clKernel.getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, sizeof(int) * (numIntParams+1), pIntParams, &err);
						std::cout << "clCreateBuffer = " << err << std::endl;
						clFloatValues = clCreateBuffer(clKernel.getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, sizeof(float) * (numFloatParams+1), pFloatParams, &err);
						std::cout << "clCreateBuffer = " << err << std::endl;
						clResults = clCreateBuffer(clKernel.getContext(), CL_MEM_WRITE_ONLY, sizeof(float) * (numx*5), 0, &err);
						std::cout << "clCreateBuffer = " << err << std::endl;

						if (pHostResults)
							delete [] pHostResults;
						if (pDoubleResults)
							delete [] pDoubleResults;

						pHostResults = new float[numx*5];
						pDoubleResults = new double[numx*5];
						
						delete [] pIntParams;
						delete [] pFloatParams;

						// Inform server that we've received the lens

						uint8_t buf[sizeof(int32_t)];
						serut::MemorySerializer mser2(0,0,buf,sizeof(int32_t));

						mser2.writeInt32(RENDERSERVER_COMMAND_RECEIVEDLENS);

						if (!connection.write(buf, sizeof(int32_t)))
						{
							std::cerr << "Error: " << std::endl;
							std::cerr << " Error writing received lens acknowledgment: " << connection.getErrorString() << std::endl; 
							return -1;
						}
					}
					else
					{
						std::cerr << "Error: " << std::endl;
						std::cerr << " Received invalid command from server" << std::endl;
						return -1;
					}
				}
			}
		}

		time_t curtime = time(0);	

		if ((curtime - connection.getLastWriteTime()) > 10) // write something every 10 seconds
		{
			uint8_t buf[sizeof(int32_t)];
			serut::MemorySerializer mser(0,0,buf,sizeof(int32_t));

			mser.writeInt32(RENDERSERVER_COMMAND_ECHO);

			if (!connection.write(buf, sizeof(int32_t)))
			{
				std::cerr << "Error: " << std::endl;
				std::cerr << " Error writing echo packet: " << connection.getErrorString() << std::endl; 
				return -1;
			}
		}

		if ((curtime - connection.getLastReadTime()) > 1200) // server died
		{
			std::cerr << "Error: " << std::endl;
			std::cerr << " Server didn't send anything for the last minute" << std::endl; 
			return -1;
		}
	}

	if (linebuffer)
		delete [] linebuffer;
	if (lens)
		delete lens;

	if (pHostResults)
		delete [] pHostResults;
	if (pDoubleResults)
		delete [] pDoubleResults;
	if (clIntValues)
		clReleaseMemObject(clIntValues);
	if (clFloatValues)
		clReleaseMemObject(clFloatValues);
	if (clResults)
		clReleaseMemObject(clResults);

	return 0;
}

int main(int argc, char *argv[])
{
#ifdef WIN32
	WSADATA dat;
	WSAStartup(MAKEWORD(2,2),&dat);
#endif // WIN32
	int status = main2(argc, argv);
#ifdef WIN32
	WSACleanup();
#endif
	return status;
}
#else

#include <iostream>

int main(void)
{
	std::cerr << "No OpenCL support enabled at compile time" << std::endl;
	return 0;
}

#endif // GRALECONFIG_SUPPORT_OPENCL
