/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "renderserver.h"
#include "gravitationallens.h"
#include <serut/memoryserializer.h>
#include <enut/tcpv4socket.h>
#include <enut/tcppacketsocket.h>
#include <enut/packet.h>
#include <enut/socketwaiter.h>
#include <signal.h>
#include <stdlib.h>
#include <iostream>
#include <string>

#include "debugnew.h"

using namespace grale;
bool endserverloop = false;

void SignalHandler(int val)
{
	endserverloop = true;
}

class ConnectionInfo;
class CurrentLensInfo;

std::list<ConnectionInfo *> connections;
ConnectionInfo *clientconn = 0;
CurrentLensInfo *lensinf = 0;

class CurrentLensInfo
{
public:
	CurrentLensInfo(const uint8_t *lensbuf,int bufsize, int numx, int numy)									
	{ 									
		lensbuffer = lensbuf; 
		lensbuffersize = bufsize; 
		lensnum++; 
		xsteps = numx; 
		ysteps = numy; 
		
		linemarkers = new bool [ysteps]; 
		
		for (int i = 0 ; i < ysteps ; i++) 
			linemarkers[i] = false; 
		calculatedlines = 0; 
		requestedlines = 0; 
		imagesize = 5*sizeof(double)*numx*numy+sizeof(int32_t); 
		
		image = new uint8_t [imagesize]; 
	
		serut::MemorySerializer mser(0, 0, image, sizeof(int32_t)); 
		mser.writeInt32(RENDERSERVER_COMMAND_IMAGEPLANE); 
	}

	~CurrentLensInfo()							
	{ 
		delete [] linemarkers; 
		delete [] lensbuffer; 
		delete [] image; 
	}
	
	const uint8_t *getLensBuffer()						{ return lensbuffer; }
	int getLensBufferSize()							{ return lensbuffersize; }
	
	int getXSteps() const							{ return xsteps; }
	int getTotalLines() const						{ return ysteps; }
	int getCalculatedLines() const						{ return calculatedlines; }

	void reenterLine(int linenr)						{ requestedlines--; linemarkers[linenr] = false; }
	int requestLine()							{ if (requestedlines == ysteps) return -1; for (int i = 0 ; i < ysteps ; i++) if (!linemarkers[i]) { linemarkers[i] = true; requestedlines++; return i; }  return -1; }
	
	void storeCalculatedLine(int l,const uint8_t *buf)			{ memcpy(image+sizeof(int32_t)+l*5*sizeof(double)*xsteps,buf,5*sizeof(double)*xsteps); calculatedlines++; }
	uint8_t *getImagePlane()						{ return image; }
	int getImagePlaneSize() const						{ return imagesize; }
		
	static int64_t getCurrentLensNumber()					{ return lensnum; }
private:
	static int64_t lensnum;

	int xsteps, ysteps;

	bool *linemarkers;
	int calculatedlines;
	int requestedlines;

	const uint8_t *lensbuffer;
	int lensbuffersize;

	uint8_t *image;
	int imagesize;
};

int64_t CurrentLensInfo::lensnum = 0;

class ConnectionInfo
{
public:
	enum ConnectionState { Unidentified, Idle, Rendering, LensTransfer };
		
	ConnectionInfo(nut::TCPPacketSocket *conn)	
	{ 
		connection = conn; 
		state = Unidentified; 
		lensnumber = -1; 
		activeline = -1; 
		solutionsent = false;
		
		std::cout << "Accepted incoming connection from " << conn->getBaseSocket()->getDestinationAddress()->getAddressString() << ":" << conn->getBaseSocket()->getDestinationPort() << std::endl;
	}
	~ConnectionInfo()							
	{ 
		std::cout << "Removing connection from " << connection->getBaseSocket()->getDestinationAddress()->getAddressString() << ":" << connection->getBaseSocket()->getDestinationPort() << std::endl;
		delete connection; 
		if (lensnumber == CurrentLensInfo::getCurrentLensNumber()) 
		{ 
			if (activeline != -1) 
				lensinf->reenterLine(activeline); 
		} 
	}
	
	nut::TCPPacketSocket *getConnection()					{ return connection; }
	void setState(ConnectionState s)					{ state = s; }
	ConnectionState getState() const					{ return state; }

	bool processPacket(const uint8_t *data, int datalength, bool *deletedata);
	bool sendData();
	bool giveFeedback(bool force = false);
	bool keepAlive();
	void sendUnavailable();
	bool writeSolution();

	int getActiveLine() const						{ return activeline; }
	int64_t getLensNumber() const						{ return lensnumber; }
	bool isSolutionSent() const						{ return solutionsent; }
private:
	nut::TCPPacketSocket *connection;
	ConnectionState state;
	int64_t lensnumber;

	int activeline;
	bool solutionsent;
};

bool ConnectionInfo::processPacket(const uint8_t *data, int datalength, bool *deletepacket)
{
	*deletepacket = true;
		
	if (datalength < sizeof(int32_t))
		return false;

	serut::MemorySerializer mser(data, datalength, 0, 0);
	int32_t cmd;
	
	mser.readInt32(&cmd);

	if (state == Unidentified)
	{
		if (cmd == RENDERSERVER_COMMAND_CLIENT)
		{
			if (clientconn != 0) // already a client connected
			{
				uint8_t buf[sizeof(int32_t)];
				serut::MemorySerializer mser(0, 0, buf, sizeof(int32_t));

				mser.writeInt32(RENDERSERVER_COMMAND_BUSY); // tell client we're busy
				connection->write(buf, sizeof(int32_t));
				
				return false;
			}
			else
			{
				if (connections.size() == 1) // this is the only connection, no renderers are present
				{
					sendUnavailable();
					return false;
				}
				else
				{
					uint8_t buf[sizeof(int32_t)];
					serut::MemorySerializer mser(0, 0, buf, sizeof(int32_t));

					mser.writeInt32(RENDERSERVER_COMMAND_ACCEPT);
					if (!connection->write(buf, sizeof(int32_t)))
						return false;

					clientconn = this;
					state = Idle;
				}
			}
		}
		else if (RENDERSERVER_COMMAND_RENDERER)
		{
			state = Idle;
		}
		else // received invalid command
			return false;
	}
	else // identified
	{
		if (this == clientconn)
		{
			if (cmd == RENDERSERVER_COMMAND_LENS) // client sent lens info to us
			{
				if (lensnumber != -1)
					return false;

				int32_t numx,numy;
				Vector2D<double> bottomleft,topright;
				double D_s, D_ds;
				
				if (!mser.readInt32(&numx))
					return false;
				if (!mser.readInt32(&numy))
					return false;
				if (numx < 2 || numx > 100000)
					return false;
				if (numy < 2 || numy > 100000)
					return false;
				
				if (!mser.readDoubles(bottomleft.getComponents(), 2))
					return false;
				if (!mser.readDoubles(topright.getComponents(), 2))
					return false;

				if (!mser.readDouble(&D_s))
					return false;
				if (!mser.readDouble(&D_ds))
					return false;

				GravitationalLens *l = 0;
				std::string errstr;
				
				if (!GravitationalLens::read(mser, &l, errstr))
					return false;
				
				delete l;
				
				if (mser.getBytesRead() != datalength)
					return false;

				// Packet is ok, store new lens info

				delete lensinf;
				lensinf = new CurrentLensInfo(data, datalength, numx, numy);
				*deletepacket = false;
				lensnumber = CurrentLensInfo::getCurrentLensNumber();
			}
			else if (cmd == RENDERSERVER_COMMAND_ECHO)
			{
				if (lensnumber == -1)
					return false;
				if (datalength != sizeof(int32_t))
					return false;
				return true;
			}
			else if (cmd == RENDERSERVER_COMMAND_ACKIMAGEPLANE)
			{
				return false; // Make sure connection is closed
			}
			else
				return false;
		}
		else // renderer connection
		{
			if (cmd == RENDERSERVER_COMMAND_ECHO)
			{
				if (datalength == sizeof(int32_t))
					return true;
				return false;
			}
			else if (cmd == RENDERSERVER_COMMAND_RESULT)
			{
				if (state != Rendering)
					return false;
			
				int replylen = datalength-sizeof(int32_t);
				if (replylen != sizeof(double)*5*lensinf->getXSteps())
					return false;
			
				lensinf->storeCalculatedLine(activeline, data+sizeof(int32_t));
				activeline = -1;
				state = Idle;
			}
			else if (cmd == RENDERSERVER_COMMAND_RECEIVEDLENS)
			{
				if (state != LensTransfer)
					return false;
				state = Idle;
			}
			else // invalid command
				return false;
		}
	}	
	return true;
}

bool ConnectionInfo::sendData()
{
	if (state != Idle)
		return true;

	// Ok, renderer is idle.
	
	if (!lensinf) // no lens info available
		return true;
	
	if (!clientconn) // there's no client connected to calculate things for
		return true;

	if (clientconn->lensnumber == -1) // client hasn't sent its lens yet
		return true;
			
	int line = lensinf->requestLine();
	
	if (line == -1) // no more lines to process
		return true;
	
	if (CurrentLensInfo::getCurrentLensNumber() != lensnumber)
	{
		// we won't send line info yet, release the line number first
		lensinf->reenterLine(line);
		
		state = LensTransfer;
		lensnumber = CurrentLensInfo::getCurrentLensNumber();

		if (!connection->write(lensinf->getLensBuffer(), lensinf->getLensBufferSize()))
			return false;
	}
	else // ok, remote side has current lens info
	{
		activeline = line;
		state = Rendering;
		
		uint8_t buf[sizeof(int32_t)+sizeof(int32_t)];
		int len = sizeof(int32_t)+sizeof(int32_t);
		serut::MemorySerializer mser(0,0,buf,len);

		mser.writeInt32(RENDERSERVER_COMMAND_RENDER);
		mser.writeInt32(line);

		if (!connection->write(buf, len))
			return false;
	}
	
	return true;
}

bool ConnectionInfo::giveFeedback(bool force)
{
	time_t curtime = time(0);
	
	if (!force)
	{
		if (curtime - connection->getLastWriteTime() < 10) // send something every ten seconds
			return true;
	}

	if (lensnumber < 0) // client hasn't sent its lens yet, nothing to give feedback about
		return true;
	
	int pct = (int)((((float)lensinf->getCalculatedLines())/((float)lensinf->getTotalLines()))*100.0);

	uint8_t data[sizeof(int32_t)*2];
	serut::MemorySerializer mser(0,0,data,sizeof(int32_t)*2);

	mser.writeInt32(RENDERSERVER_COMMAND_PCT);
	mser.writeInt32(pct);
	
	if (!connection->write(data, sizeof(int32_t)*2))
		return false;
	
	return true;
}

bool ConnectionInfo::keepAlive()
{
	time_t curtime = time(0);
	
	if ((curtime - connection->getLastWriteTime()) < 10) // make sure something is sent every ten seconds
		return true;
		
	uint8_t buf[sizeof(int32_t)];
	serut::MemorySerializer mser(0,0,buf,sizeof(int32_t));
	
	mser.writeInt32(RENDERSERVER_COMMAND_ECHO);
	if (!connection->write(buf,sizeof(int32_t)))
		return false;
	
	return true;
}

void ConnectionInfo::sendUnavailable()
{
	uint8_t buf[sizeof(int32_t)];
	serut::MemorySerializer mser(0,0,buf,sizeof(int32_t));

	mser.writeInt32(RENDERSERVER_COMMAND_UNAVAILABLE);
	connection->write(buf,sizeof(int32_t));
}
	
bool ConnectionInfo::writeSolution()
{
	solutionsent = true;
	if (!connection->write(lensinf->getImagePlane(),lensinf->getImagePlaneSize()))
		return false;
	return true;
}

int main2(int argc,char *argv[])
{
	if (argc != 2)
	{
		std::cerr << "Usage: " << std::endl;
		std::cerr << std::string(argv[0]) << " portnumber" << std::endl; 
		return -1;
	}

	nut::TCPv4Socket listenSocket;

	if (!listenSocket.create(atoi(argv[1])))
	{
		std::cerr << "Error creating server at port " << atoi(argv[1]) << ":" << std::endl;
		std::cerr << listenSocket.getErrorString() << std::endl;
		return -1;
	}

	if (!listenSocket.listen(32))
	{
		std::cerr << "Couldn't set socket in listen mode:" << std::endl;
		std::cerr << listenSocket.getErrorString() << std::endl;
		return -1;
	}

	std::cout << "Listen socket created at port " << listenSocket.getLocalPortNumber() << std::endl;

#ifndef WIN32
	signal(SIGTERM,SignalHandler);
	signal(SIGHUP,SignalHandler);
	signal(SIGQUIT,SignalHandler);
	signal(SIGABRT,SignalHandler);
	signal(SIGINT,SignalHandler);
#endif // WIN32
	
	while (!endserverloop)
	{
//		std::cout << "Start of server loop" << std::endl;

		nut::SocketWaiter waiter;
		std::list<ConnectionInfo *>::iterator it;

		waiter.addSocket(listenSocket);

		for (it = connections.begin() ; it != connections.end() ; it++)
		{
			nut::TCPPacketSocket *pSock = (*it)->getConnection();
			waiter.addSocket(*pSock);
		}

//		std::cout << "Calling select" << std::endl;

		if (!waiter.wait(1, 0))
		{
			endserverloop = true;
			continue;
		}

//		std::cout << "Checking existing connections" << std::endl;
		
		it = connections.begin();
		while (it != connections.end())
		{
			ConnectionInfo *pConnInf = *it;
			if (pConnInf->getConnection()->isDataAvailable())
			{
				bool closed = false;
				size_t length = 0;
				
				pConnInf->getConnection()->getAvailableDataLength(length);
				if (length == 0)
					closed = true;
				else
				{
					if (!pConnInf->getConnection()->poll())
						closed = true;
				}
				
				if (closed)
				{
					if (pConnInf == clientconn)
						clientconn = 0;
					
					it = connections.erase(it);
					delete pConnInf;
				}
				else // connection not closed, process received packet (if available)
				{
					bool goterror = false;
					nut::Packet packet;
					
					while (!goterror && (pConnInf->getConnection()->read(packet)))
					{
						bool deletedata;
						size_t length = 0;
						uint8_t *pData = packet.extractData(length);

						if (!pConnInf->processPacket(pData, length, &deletedata))
							goterror = true;
						if (deletedata)
							delete [] pData;
					}

					if (goterror)
					{
						// some inconsistency was detected, remove connection
		
						if (pConnInf == clientconn)
							clientconn = 0;
						
						it = connections.erase(it);
						delete pConnInf;
					}
					else
						it++;
				}
			}
			else
				it++;
		}	
	
//		std::cout << "Checking incoming connections" << std::endl;
		
		if (listenSocket.isDataAvailable())
		{
			// Incoming connection: accept it

			nut::TCPSocket *pSock;

			if (listenSocket.accept(&pSock))
			{
				nut::TCPPacketSocket *pConn = new nut::TCPPacketSocket(pSock, true, false, RENDERSERVER_MAXPACKSIZE, sizeof(uint32_t), RENDERSERVER_PACKID);
				ConnectionInfo *inf = new ConnectionInfo(pConn);
				connections.push_back(inf);
			}
		}

		// Processed all incoming stuff, now we'll check if we can send some data

		it = connections.begin();
		while (it != connections.end())
		{
			ConnectionInfo *inf = (*it);

			if (inf != clientconn)
			{
				if (!inf->sendData())
				{
					it = connections.erase(it);
					delete inf;
				}
				else
					it++;
			}
			else
			{
				if (!inf->giveFeedback())
				{
					it = connections.erase(it);
					delete inf;
					clientconn = 0;
				}
				else
					it++;
			}
		}

		// Keep connections alive
		
		it = connections.begin();
		while (it != connections.end())
		{
			ConnectionInfo *inf = (*it);

			if (!inf->keepAlive())
			{
				it = connections.erase(it);
				if (inf == clientconn)
					clientconn = 0;
				delete inf;
			}
			else
				it++;
		}

		// Timeout old connections
		
		time_t curtime = time(0);
		
		it = connections.begin();
		while (it != connections.end())
		{
			ConnectionInfo *inf = (*it);

			if ((curtime - inf->getConnection()->getLastReadTime()) > 1200) // wait at most twenty minutes
			{
				it = connections.erase(it);
				if (inf == clientconn)
					clientconn = 0;
				delete inf;
			}
			else
				it++;
		}

		// Check if the entire image plane was calculated

		if (clientconn && lensinf && clientconn->getLensNumber() == CurrentLensInfo::getCurrentLensNumber() && !clientconn->isSolutionSent() && (lensinf->getCalculatedLines() == lensinf->getTotalLines()))
		{
			// write result
			
			clientconn->giveFeedback(true);
			if (!clientconn->writeSolution())
			{
				// remove from connections list
			
				it = connections.begin();
				while (it != connections.end())
				{
					ConnectionInfo *inf = (*it);
	
					if (inf == clientconn)
						it = connections.erase(it);
					else
						it++;
				}
				delete clientconn;
				clientconn = 0;
			}
		}

		// Check if renderers are available to help client
		if (clientconn)
		{
			if (connections.size() == 1)
			{
				// This means that we can't help the client just now, report this
				
				clientconn->sendUnavailable();
				delete clientconn;
				clientconn = 0;
				connections.clear();
			}
		}
	}

	while (!connections.empty()) // clientconn is also in here
	{
		delete *(connections.begin());
		connections.pop_front();
	}

	if (lensinf)
		delete lensinf;

	std::cout << "Exiting" << std::endl;

	return 0;
}

int main(int argc, char *argv[])
{
#ifdef WIN32
	WSADATA dat;
	WSAStartup(MAKEWORD(2,2),&dat);
#endif // WIN32
	int status = main2(argc, argv);
#ifdef WIN32
	WSACleanup();
#endif
	return status;
}