/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "gridlensinversiongafactorynew.h"
#include "gridlensinversiongafactoryparamsnew.h"
#include "gridlensinversiongenomenew.h"
#include "compositelens.h"
#include "plummerlens.h"
#include "squarelens.h"
#include "gausslens.h"
#include "masssheetlens.h"
#include "lensfitnessobject.h"
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_errno.h>

#include <iostream>

#include "gravitationallens.h"
#include <sys/stat.h>
#include <sys/types.h>

#ifdef WIN32
#include <windows.h>
#endif // WIN32

#include "debugnew.h"

namespace grale
{

GridLensInversionGAFactoryNew::GridLensInversionGAFactoryNew()
{
	m_pCurrentParams = 0;
	m_numMasses = 0;
	m_maxGenerations = 0;
	m_allowNegativeValues = false;
	m_useGenomeSheet = false;
	m_sheetScale = 0;
	m_pDeflectionMatrix = 0;
	m_pShortBPMatrix = 0;
	m_pTotalBPMatrix = 0;
	m_pFitnessObject = 0;

	m_showEvolution = false;
	m_saveFirstGenome = false;
}

GridLensInversionGAFactoryNew::~GridLensInversionGAFactoryNew()
{
	if (m_pCurrentParams)
		delete m_pCurrentParams;
	for (int i = 0 ; i < m_basisLenses.size() ; i++)
		delete m_basisLenses[i].first;
	if (m_pDeflectionMatrix)
		delete m_pDeflectionMatrix;
	if (m_pShortBPMatrix)
	{
		if (m_pShortBPMatrix != m_pTotalBPMatrix)
			delete m_pShortBPMatrix;
	}
	if (m_pTotalBPMatrix)
		delete m_pTotalBPMatrix;
	if (m_pFitnessObject)
		delete m_pFitnessObject;
}

mogal::GAFactoryParams *GridLensInversionGAFactoryNew::createParamsInstance() const
{
	return new GridLensInversionGAFactoryParamsNew();
}

const mogal::GAFactoryParams *GridLensInversionGAFactoryNew::getCurrentParameters() const
{
	return m_pCurrentParams;
}

bool GridLensInversionGAFactoryNew::init(const mogal::GAFactoryParams *pParams)
{
	if (m_pCurrentParams != 0)
	{
		setErrorString("Already initialized");
		return false;
	}

	if (pParams == 0)
	{
		setErrorString("Specified parameters can't be null");
		return false;
	}
	
	const LensInversionGAFactoryParams *pParams0 = (const LensInversionGAFactoryParams *)pParams;
	if (pParams0->getInversionType() != LensInversionGAFactoryParams::GridInversionNew)
	{
		setErrorString("This algorithm is not a new grid based one");
		return false;
	}

	const GridLensInversionGAFactoryParamsNew *pParams2 = (const GridLensInversionGAFactoryParamsNew *)pParams;
	
	m_pCurrentParams = pParams2->createCopy(true);
	m_numMasses = m_pCurrentParams->getGridSquares().size();
	m_showEvolution = m_pCurrentParams->getShowEvolution();
	m_saveFirstGenome = m_pCurrentParams->getSaveFirstGenome();
	
	// Create the basis functions

	const std::list<GridSquare> &squares = m_pCurrentParams->getGridSquares();
	std::vector<std::pair<GravitationalLens *, Vector2D<double> > > basisLenses;
	std::vector<double> massWeights(m_numMasses);

	for (int i = 0 ; i < m_numMasses ; i++)
		massWeights[i] = 1.0/(double)m_numMasses;

	double massScale = m_pCurrentParams->getInitialMassScale();
	double D_d = m_pCurrentParams->getLensDistance();
	double minSize;

	calculateMassRegion(squares, minSize);

	if (!createBasisLenses(D_d, squares, massWeights, massScale, basisLenses))
	{
		delete m_pCurrentParams;
		m_pCurrentParams = 0;
		return false;
	}

	if (!recalculateMassWeights(basisLenses, massWeights, squares, massScale))
	{
		for (int i = 0 ; i < basisLenses.size() ; i++)
			delete basisLenses[i].first;
		basisLenses.clear();
		delete m_pCurrentParams;
		m_pCurrentParams = 0;
		return false;
	}

	for (int i = 0 ; i < basisLenses.size() ; i++)
		delete basisLenses[i].first;
	basisLenses.clear();

	// Use the new weights to create the actual basis functions
	if (!createBasisLenses(D_d, squares, massWeights, massScale, basisLenses))
	{
		delete m_pCurrentParams;
		m_pCurrentParams = 0;
		return false;
	}

	bool useSheet = false;

	if (pParams2->getMassSheetSearchType() != GridLensInversionGAFactoryParamsNew::None)
	{
		useSheet = true;
		// TODO: what is a good value here?
		m_sheetScale = 1.1;
		m_useGenomeSheet = true;
	}
	else
	{
		useSheet = false;
		m_sheetScale = 0;
		m_useGenomeSheet = false;
	}

	// perform sub-initialization
	if (!localSubInit(m_pCurrentParams->getLensRedshift(), m_pCurrentParams->getImages(), basisLenses, m_pCurrentParams->getBaseLens(), useSheet))
	{
		// Error string is set in subInit
		for (int i = 0 ; i < basisLenses.size() ; i++)
			delete basisLenses[i].first;
		basisLenses.clear();
		delete m_pCurrentParams;
		m_pCurrentParams = 0;
		return false;
	}
	
	m_basisLenses = basisLenses;
	m_maxGenerations = m_pCurrentParams->getMaximumNumberOfGenerations();
	m_allowNegativeValues = m_pCurrentParams->allowNegativeValues();

	double angularScale = m_pDeflectionMatrix->getAngularScale();

	m_bottomLeftFloat = Vector2D<float>((float)(m_bottomLeft.getX()/angularScale), (float)(m_bottomLeft.getY()/angularScale) );
	m_topRightFloat = Vector2D<float>((float)(m_topRight.getX()/angularScale), (float)(m_topRight.getY()/angularScale) );

	m_basisCenters.resize(m_basisLenses.size());

	for (int i = 0 ; i < basisLenses.size() ; i++)
		m_basisCenters[i] = Vector2D<float>( (float) (m_basisLenses[i].second.getX()/angularScale),
				                    (float) (m_basisLenses[i].second.getY()/angularScale) );

	Vector2D<float> diagonal = m_topRightFloat - m_bottomLeftFloat;
	
	// TODO: remove this?
	m_minMutationWidth = (float)(minSize/angularScale);
	m_maxMutationWidth = diagonal.getLength();

	std::cout << "min: " << m_minMutationWidth << std::endl;
	std::cout << "max: " << m_maxMutationWidth << std::endl;

	//
	// TODO: for debugging
	//
#if 0
	std::vector<float> masses(m_numMasses);
	double totalMass;
	std::string errStr;

	for (int i = 0 ; i < m_numMasses ; i++)
		masses[i] = 1.0f;

	GravitationalLens *pLens = createLens(masses, 0, 1.0f, &totalMass, errStr);
	pLens->save("testlens.lensdata");

	exit(-1);
#endif

	return true;
}

mogal::Genome *GridLensInversionGAFactoryNew::createNewGenome() const
{
	return new GridLensInversionGenomeNew((GridLensInversionGAFactoryNew *)this, m_numMasses, m_useGenomeSheet);
}

bool GridLensInversionGAFactoryNew::writeGenome(serut::SerializationInterface &si, const mogal::Genome *pGenome) const
{
	const GridLensInversionGenomeNew *pGenome2 = (const GridLensInversionGenomeNew *)pGenome;

	if (!si.writeFloats(pGenome2->getMasses()))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeFloat(pGenome2->getSheetValue()))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	return true;
}

bool GridLensInversionGAFactoryNew::readGenome(serut::SerializationInterface &si, mogal::Genome **pGenome) const
{
	std::vector<float> masses(m_numMasses);
	float sheetValue;

	if (!si.readFloats(masses))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.readFloat(&sheetValue))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	*pGenome = new GridLensInversionGenomeNew((GridLensInversionGAFactoryNew *)this, masses, sheetValue);
	return true;
}

bool GridLensInversionGAFactoryNew::writeGenomeFitness(serut::SerializationInterface &si, const mogal::Genome *pGenome) const
{
	const GridLensInversionGenomeNew *pGenome2 = (const GridLensInversionGenomeNew *)pGenome;
	int num = getNumberOfFitnessComponents();
	const float *pValues = pGenome2->getFitnessValues();

	if (!si.writeFloats(pValues, num))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeFloat(pGenome2->getScaleFactor()))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	return true;
}

bool GridLensInversionGAFactoryNew::readGenomeFitness(serut::SerializationInterface &si, mogal::Genome *pGenome) const
{
	GridLensInversionGenomeNew *pGenome2 = (GridLensInversionGenomeNew *)pGenome;
	int num = getNumberOfFitnessComponents();
	float x[GRIDLENSINVERSIONGENOMENEW_MAXFITNESSCOMP+1];

	if (!si.readFloats(x, num+1))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	pGenome2->setFitnessValues(x);
	pGenome2->setScaleFactor(x[num]);
	return true;
}

bool GridLensInversionGAFactoryNew::writeCommonGenerationInfo(serut::SerializationInterface &si) const
{
	return true;
}

bool GridLensInversionGAFactoryNew::readCommonGenerationInfo(serut::SerializationInterface &si)
{
	return true;
}

GravitationalLens *GridLensInversionGAFactoryNew::createLens(const std::vector<float> &masses, float sheetValue, float scaleFactor, 
                                                              double *pTotalMass, std::string &errStr) const
{
	CompositeLensParams lensParams;
	CompositeLens *pLens;

	for (int i = 0 ; i < m_basisLenses.size() ; i++)
		lensParams.addLens((double)masses[i]*(double)scaleFactor, m_basisLenses[i].second, 0, *(m_basisLenses[i].first));

	if (m_sheetScale != 0)
	{
		double D_d = m_basisLenses[0].first->getLensDistance();
		MassSheetLensParams sheetParams(((SPEED_C*SPEED_C)/(4.0*CONST_PI*CONST_G*D_d))*sheetValue*m_sheetScale);
		MassSheetLens sheetLens;

		if (!sheetLens.init(D_d, &sheetParams))
		{
			setErrorString(std::string("Can't initialize mass sheet component: ") + sheetLens.getErrorString());
			return 0;
		}

		lensParams.addLens(1.0, Vector2D<double>(0,0), 0, sheetLens);
	}

	pLens = new CompositeLens();
	pLens->init(m_basisLenses[0].first->getLensDistance(), &lensParams);

	double totalMass = 0; // TODO: is without the mass sheet

	for (int i = 0 ; i < m_basisLenses.size() ; i++)
	{
		if (m_basisLenses[i].first->getLensType() == GravitationalLens::Plummer)
		{
			const PlummerLensParams *pParams = (const PlummerLensParams *)m_basisLenses[i].first->getLensParameters();

			totalMass += pParams->getLensMass()*(double)scaleFactor*(double)masses[i];
		}
		else if (m_basisLenses[i].first->getLensType() == GravitationalLens::Gaussian)
		{
			const GaussLensParams *pParams = (const GaussLensParams *)m_basisLenses[i].first->getLensParameters();

			totalMass += pParams->getMass()*(double)scaleFactor*(double)masses[i];
		}
		else if (m_basisLenses[i].first->getLensType() == GravitationalLens::Square)
		{
			const SquareLensParams *pParams = (const SquareLensParams *)m_basisLenses[i].first->getLensParameters();

			totalMass += pParams->getLensMass()*(double)scaleFactor*(double)masses[i];
		}
	}

	// TODO:
//	std::cout << "Total mass is: " << totalMass/(MASS_SOLAR) << std::endl;

	*pTotalMass = totalMass;

	return pLens;
}

void GridLensInversionGAFactoryNew::onSortedPopulation(const std::vector<mogal::GenomeWrapper> &population)
{
	if (m_showEvolution)
	{
		char str[1024];
		int generation = getCurrentAlgorithm()->getCurrentGeneration();

		sprintf(str,"g_%d",generation);
		
#ifndef WIN32
		if(mkdir(str,S_IRWXU) != 0)
			return;
#else
		if(!CreateDirectory(str,0))
			return;
#endif // WIN32
		
		for (int i = 0 ; i < (int)population.size() ; i++)
		{
			char str[1024];
			sprintf(str,"g_%d/p_%d_%d_%d.lensdata",generation,i,population[i].getParent1(),population[i].getParent2());

			const LensInversionGenomeBase *g = (const LensInversionGenomeBase *)population[i].getGenome();

			double m = 0;
			std::string errstr;
			GravitationalLens *lens = g->createLens(&m,errstr);
			lens->save(str);
			delete lens;
		}
	}

	if (m_saveFirstGenome)
	{
		int generation = getCurrentAlgorithm()->getCurrentGeneration();
		char str[1024];

		if (generation == 0)
#ifndef WIN32
			mkdir("bestgenomes",S_IRWXU);
#else
			CreateDirectory("bestgenomes", 0);
#endif // WIN32
	
		sprintf(str,"bestgenomes/gen_%d.lensdata",generation);

		const LensInversionGenomeBase *g = (const LensInversionGenomeBase *)population[0].getGenome();

		double m = 0;
		std::string errstr;
		GravitationalLens *lens = g->createLens(&m,errstr);
		lens->save(str);
		delete lens;
	}
}

bool GridLensInversionGAFactoryNew::localSubInit(double z_d, const std::list<ImagesDataExtended *> &images, 
	                  const std::vector<std::pair<GravitationalLens *, Vector2D<double> > > &basisLenses,
			  const GravitationalLens *pBaseLens, bool useSheet)
{
	std::list<ImagesDataExtended *> reducedImages = images;
	std::list<ImagesDataExtended *> shortImages;

	m_pFitnessObject = createFitnessObject();
	if (m_pFitnessObject == 0)
		return false; // error string should be set in the createFitnessObject function

	if (!m_pFitnessObject->init(z_d, reducedImages, shortImages, 0)) // TODO: specify actual parameters
	{
		setErrorString(m_pFitnessObject->getErrorString());
		delete m_pFitnessObject;
		m_pFitnessObject = 0;
		return false;
	}

	std::vector<ImagesDataExtended *> reducedImagesVector, shortImagesVector;
	std::list<ImagesDataExtended *>::const_iterator it;

	for (it = reducedImages.begin() ; it != reducedImages.end() ; it++)
		reducedImagesVector.push_back(*it);

	for (it = shortImages.begin() ; it != shortImages.end() ; it++)
		shortImagesVector.push_back(*it);
	
	// TODO: this has to be done in a subclass!
	//setNumberOfFitnessComponents(m_pFitnessObject->getNumberOfFitnessComponents());

	if (!subInit(m_pFitnessObject))
	{
		delete m_pFitnessObject;
		m_pFitnessObject = 0;
		return false;
	}

	m_pDeflectionMatrix = new DeflectionMatrix();
	if (!m_pDeflectionMatrix->startInit())
	{
		setErrorString(m_pDeflectionMatrix->getErrorString());
		delete m_pDeflectionMatrix;
		delete m_pFitnessObject;
		m_pDeflectionMatrix = 0;
		m_pFitnessObject = 0;
		return false;
	}

	bool totalStoreIntens, totalStoreTimeDelay, totalStoreShearInfo;
	std::vector<bool> totalDeflectionFlags, totalDerivativeFlags, totalPotentialFlags;

	m_pFitnessObject->getTotalCalcFlags(totalDeflectionFlags, totalDerivativeFlags, totalPotentialFlags);
	m_pFitnessObject->getTotalStoreFlags(&totalStoreIntens, &totalStoreTimeDelay, &totalStoreShearInfo);

	m_pTotalBPMatrix = new BackProjectMatrixNew();
	if (!m_pTotalBPMatrix->startInit(z_d, basisLenses[0].first->getLensDistance(), m_pDeflectionMatrix,
				         reducedImagesVector, totalDeflectionFlags, totalDerivativeFlags, 
					 totalPotentialFlags, pBaseLens, totalStoreIntens, totalStoreTimeDelay,
					 totalStoreShearInfo,
					 useSheet))
	{
		setErrorString(m_pTotalBPMatrix->getErrorString());
		delete m_pTotalBPMatrix;
		delete m_pDeflectionMatrix;
		delete m_pFitnessObject;
		m_pDeflectionMatrix = 0;
		m_pFitnessObject = 0;
		m_pTotalBPMatrix = 0;
		return false;
	}

	if (shortImagesVector.size() > 0) // Ok, can improve speed in mass-scale calculation
	{
		bool shortStoreIntens, shortStoreTimeDelay, shortStoreShearInfo;
		std::vector<bool> shortDeflectionFlags, shortDerivativeFlags, shortPotentialFlags;

		m_pFitnessObject->getShortCalcFlags(shortDeflectionFlags, shortDerivativeFlags, shortPotentialFlags);
		m_pFitnessObject->getShortStoreFlags(&shortStoreIntens, &shortStoreTimeDelay, &shortStoreShearInfo);

		m_pShortBPMatrix = new BackProjectMatrixNew();
		if (!m_pShortBPMatrix->startInit(z_d, basisLenses[0].first->getLensDistance(), m_pDeflectionMatrix,
						 shortImagesVector, shortDeflectionFlags, shortDerivativeFlags, 
						 shortPotentialFlags, pBaseLens, shortStoreIntens, shortStoreTimeDelay,
						 shortStoreShearInfo,
						 useSheet))
		{
			setErrorString(m_pShortBPMatrix->getErrorString());
			delete m_pShortBPMatrix;
			delete m_pTotalBPMatrix;
			delete m_pDeflectionMatrix;
			delete m_pFitnessObject;
			m_pDeflectionMatrix = 0;
			m_pFitnessObject = 0;
			m_pTotalBPMatrix = 0;
			m_pShortBPMatrix = 0;
			return false;
		}

	}
	else
		m_pShortBPMatrix = m_pTotalBPMatrix;

	bool err = false;

	if (!m_pDeflectionMatrix->endInit(basisLenses))
	{
		setErrorString(m_pDeflectionMatrix->getErrorString());
		err = true;
	}
	if (!err)
	{
		if (!m_pTotalBPMatrix->endInit())
		{
			setErrorString(m_pTotalBPMatrix->getErrorString());
			err = true;
		}
	}
	if (!err)
	{
		if (m_pShortBPMatrix != m_pTotalBPMatrix)
		{
			if (!m_pShortBPMatrix->endInit())
			{
				setErrorString(m_pShortBPMatrix->getErrorString());
				err = true;
			}
		}
	}

	if (err)
	{
		if (m_pShortBPMatrix != m_pTotalBPMatrix)
		{
			delete m_pShortBPMatrix;
			m_pShortBPMatrix = 0;
		}
		delete m_pTotalBPMatrix;
		delete m_pDeflectionMatrix;
		delete m_pFitnessObject;
		m_pDeflectionMatrix = 0;
		m_pFitnessObject = 0;
		m_pTotalBPMatrix = 0;
		return false;
	}

	m_pFitnessObject->postInit(reducedImages, shortImages, m_pDeflectionMatrix->getAngularScale());

	return true;
}

bool GridLensInversionGAFactoryNew::createBasisLenses(double D_d, const std::list<GridSquare> &squares, 
			       const std::vector<double> &massWeights, double massScale,
			       std::vector<std::pair<GravitationalLens *, Vector2D<double> > > &basisLenses)
{
	std::list<GridSquare>::const_iterator squareIt;
	int squareNumber = 0;

	for (squareIt = squares.begin() ; squareIt != squares.end() ; squareIt++, squareNumber++)
	{
		double gridSize = (*squareIt).getSize();
		Vector2D<double> gridCenter = (*squareIt).getCenter();
		GravitationalLens *pLens = 0;
		bool returnValue;
	
		if (m_pCurrentParams->getBasisFunctionType() == GridLensInversionGAFactoryParamsNew::PlummerBasis)
		{
			PlummerLensParams lensParams(massScale * massWeights[squareNumber], gridSize);

			pLens = new PlummerLens();
			returnValue = pLens->init(D_d, &lensParams);
		}
		else if (m_pCurrentParams->getBasisFunctionType() == GridLensInversionGAFactoryParamsNew::GaussBasis)
		{
			GaussLensParams lensParams(massScale * massWeights[squareNumber], gridSize);

			pLens = new GaussLens();
			returnValue = pLens->init(D_d, &lensParams);
		}
		else // Squares
		{
			SquareLensParams lensParams(massScale * massWeights[squareNumber], gridSize);

			pLens = new SquareLens();
			returnValue = pLens->init(D_d, &lensParams);
		}

		if (!returnValue)
		{
			setErrorString(std::string("Couldn't initialize basis lens: ") + pLens->getErrorString());
			for (int i = 0 ; i < basisLenses.size() ; i++)
				delete basisLenses[i].first;
			basisLenses.clear();
			delete pLens;
			return false;
		}

		basisLenses.push_back(std::pair<GravitationalLens *, Vector2D<double> >(pLens, gridCenter));
	}
	return true;
}

bool GridLensInversionGAFactoryNew::recalculateMassWeights(
		                    const std::vector<std::pair<GravitationalLens *, Vector2D<double> > > &basisLenses, 
			            std::vector<double> &massWeights, const std::list<GridSquare> &squares, 
				    double massScale)
{
	int numBasisFunctions = basisLenses.size();
	int densSamp = 100; // TODO: auto-adjust this?
	int numDensPoints = densSamp * densSamp; 

	double maxX = m_topRight.getX();
	double maxY = m_topRight.getY();
	double minX = m_bottomLeft.getX();
	double minY = m_bottomLeft.getY();

	std::vector<double> densities(numDensPoints);
	std::vector<double> trialWeights(numBasisFunctions);
	std::vector<std::vector<double> > matrix(numDensPoints);

	for (int i = 0 ; i < numDensPoints ; i++)
	{
		matrix[i].resize(numBasisFunctions);

		Vector2D<double> pos = Vector2D<double>((maxX-minX)*((double)(i%densSamp)/(double)(densSamp-1)) + minX, 
				                        (maxY-minY)*((double)(i/densSamp)/(double)(densSamp-1)) + minY);

		for (int j = 0 ; j < numBasisFunctions ; j++)
		{
			Vector2D<double> diff = pos - basisLenses[j].second;
			double dens = basisLenses[j].first->getSurfaceMassDensity(diff);

			matrix[i][j] = dens;	
		}
	}

	double bestFlatness = -1;
	double bestPower = -1;
	
	for (int i = 0 ; i < 100 ; i++)
	{
		double power = 1.5+(((double)i)/100.0)*2.5;
		double sum = 0;
	
		std::list<GridSquare>::const_iterator it = squares.begin();

		for (int i = 0 ; i < numBasisFunctions ; i++, it++)
		{
			double size = (*it).getSize();

			trialWeights[i] = POW(size, power);
			sum += trialWeights[i];
		}
		for (int i = 0 ; i < numBasisFunctions ; i++, it++)
			trialWeights[i] /= sum;
#if 0
		double avg = 0;

		for (int i = 0 ; i < numDensPoints ; i++)
		{
			double dens = 0;

			for (int j = 0 ; j < numBasisFunctions ; j++)
				dens += matrix[i][j]*trialWeights[j];

			densities[i] = dens;
			avg += dens;
		}

		avg /= (double)numDensPoints;

		double dispersion = 0;
		for (int i = 0 ; i < numDensPoints ; i++)
			dispersion += (densities[i]-avg)*(densities[i]-avg);
		dispersion /= numDensPoints;
#else
		for (int i = 0 ; i < numDensPoints ; i++)
		{
			double dens = 0;

			for (int j = 0 ; j < numBasisFunctions ; j++)
				dens += matrix[i][j]*trialWeights[j];

			densities[i] = dens;
		}

		double enumSum = 0;
		double denomSum = 0;
		double targetDens = 1.0;

		for (int i = 0 ; i < numDensPoints ; i++)
		{
			enumSum += targetDens*densities[i];
			denomSum += densities[i]*densities[i];
		}

		double lambda = enumSum/denomSum;
		double dispersion = 0;

		for (int i = 0 ; i < numDensPoints ; i++)
			dispersion += (targetDens-lambda*densities[i])*(targetDens-lambda*densities[i]);
		dispersion /= (double)numDensPoints;
#endif
		if (bestFlatness < 0 || dispersion < bestFlatness)
		{
			bestPower = power;
			bestFlatness = dispersion;
		}
	}
	
	// Store final weights
	{
		double power = bestPower;
		double sum = 0;
	
		std::list<GridSquare>::const_iterator it = squares.begin();

		for (int i = 0 ; i < numBasisFunctions ; i++, it++)
		{
			double size = (*it).getSize();

			trialWeights[i] = POW(size, power);
			sum += trialWeights[i];
		}
		for (int i = 0 ; i < numBasisFunctions ; i++)
			trialWeights[i] /= sum;

		// TODO: for testing
//		for (int i = 0 ; i < numPoints ; i++)
//			trialWeights[i] = 1.0/(double)numPoints;

		massWeights = trialWeights;
	}

//	for (int i = 0 ; i < numPoints ; i++)
//		std::cout << i << " " << massWeights[i] << std::endl;

//	std::cout << "Best power = " << bestPower << std::endl;

	return true;
}

void GridLensInversionGAFactoryNew::calculateMassRegion(const std::list<GridSquare> &squares, double &minSize)
{
	double maxX = -1e30;
	double maxY = -1e30;
	double minX = 1e30;
	double minY = 1e30;
	double minimalWidth = 1e30;
	int numBasisFunctions = squares.size();

	std::list<GridSquare>::const_iterator it = squares.begin();

	for (int i = 0 ; i < numBasisFunctions ; i++, it++)
	{
		GridSquare square = *it;
		double x = square.getCenter().getX();
		double y = square.getCenter().getY();

		if (maxX < x)
			maxX = x;
		if (maxY < y)
			maxY = y;
		if (minX > x)
			minX = x;
		if (minY > y)
			minY = y;

		if (square.getSize() < minimalWidth)
			minimalWidth = square.getSize();
	}

	m_bottomLeft = Vector2D<double>(minX, minY);
	m_topRight = Vector2D<double>(maxX, maxY);

	minSize = minimalWidth;
}

} // end namespace
