/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "gridlensinversiongafactoryparamsnew.h"
#include "imagesdataextended.h"
#include "gravitationallens.h"
#include <serut/memoryserializer.h>
#include <serut/dummyserializer.h>
#include <vector>

#include <iostream>

#include "debugnew.h"

namespace grale
{

GridLensInversionGAFactoryParamsNew::GridLensInversionGAFactoryParamsNew() : LensInversionGAFactoryParams(LensInversionGAFactoryParams::GridInversionNew)
{
	m_maxGenerations = 0;
	m_deleteImages = false;
	m_lensDistance = 0;
	m_lensRedshift = 0;
	m_initialMassScale = 0;
	m_allowNegative = false;
	m_basisFunctionType = GridLensInversionGAFactoryParamsNew::PlummerBasis;
	m_pBaseLens = 0;
	m_massSheetSearchType = GridLensInversionGAFactoryParamsNew::None;
	m_showEvolution = false;
	m_saveFirstGenome = false;
}

GridLensInversionGAFactoryParamsNew::GridLensInversionGAFactoryParamsNew(
					 int maxGenerations,
					 const std::list<ImagesDataExtended *> &images, 
			                 const std::list<GridSquare> &gridSquares, 
					 double lensDistance,
					 double lensRedshift,
					 double initialMassScale,
					 bool copyImages,
					 bool allowNegativeValues,
					 const GravitationalLens *pBaseLens,
					 BasisFunctionType basisFunction,
					 MassSheetSearchType sheetSearchType,
					 bool showEvolution,
					 bool saveFirstGenome) : LensInversionGAFactoryParams(LensInversionGAFactoryParams::GridInversionNew)
{
	m_maxGenerations = maxGenerations;

	m_lensDistance = lensDistance;
	m_lensRedshift = lensRedshift;
	m_initialMassScale = initialMassScale;
	
	if (copyImages)
	{
		m_deleteImages = true;

		std::list<ImagesDataExtended *>::const_iterator it;

		for (it = images.begin() ; it != images.end() ; it++)
		{
			ImagesDataExtended *pImg = *it;

			m_images.push_back(new ImagesDataExtended(*pImg));
		}
	}
	else
	{
		m_images = images;
		m_deleteImages = false;
	}

	m_gridSquares = gridSquares;
	m_basisFunctionType = basisFunction;
	m_allowNegative = allowNegativeValues;
	m_pBaseLens = 0;

	if (pBaseLens)
		m_pBaseLens = pBaseLens->createCopy(); // TODO: check if this doens't cause dll memory problem

	m_massSheetSearchType = sheetSearchType;
	m_showEvolution = showEvolution;
	m_saveFirstGenome = saveFirstGenome;
}

GridLensInversionGAFactoryParamsNew::~GridLensInversionGAFactoryParamsNew()
{
	if (m_deleteImages)
	{
		while (!m_images.empty())
		{
			ImagesDataExtended *pImg = *(m_images.begin());

			delete pImg;

			m_images.pop_front();
		}
	}

	if (m_pBaseLens)
		delete m_pBaseLens;
}

#define BASISFUNCTION_PLUMMER 1
#define BASISFUNCTION_SQUARE  2
#define BASISFUNCTION_GAUSS   3

#define SHEETSEARCHTYPE_NONE  	0
#define SHEETSEARCHTYPE_GENOME	1

bool GridLensInversionGAFactoryParamsNew::write(serut::SerializationInterface &si) const
{
	if (!si.writeInt32(m_maxGenerations))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDouble(m_lensDistance))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDouble(m_lensRedshift))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDouble(m_initialMassScale))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t sizes[2];

	sizes[0] = m_images.size();
	sizes[1] = m_gridSquares.size();

	if (!si.writeInt32s(sizes, 2))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	std::list<ImagesDataExtended *>::const_iterator it;

	for (it = m_images.begin() ; it != m_images.end() ; it++)
	{
		ImagesDataExtended *pImg = *it;

		if (!pImg->write(si))
		{
			setErrorString(pImg->getErrorString());
			return false;
		}
	}

	std::list<GridSquare>::const_iterator it2;

	for (it2 = m_gridSquares.begin() ; it2 != m_gridSquares.end() ; it2++)
	{
		GridSquare square = *it2;
		double params[3];

		params[0] = square.getSize();
		params[1] = square.getCenter().getX();
		params[2] = square.getCenter().getY();
	
		if (!si.writeDoubles(params, 3))
		{
			setErrorString(si.getErrorString());
			return false;
		}
	}

	int32_t b = 0;

	switch(m_basisFunctionType)
	{
	case PlummerBasis:
		b = BASISFUNCTION_PLUMMER;
		break;
	case SquareBasis:
		b = BASISFUNCTION_SQUARE;
		break;
	case GaussBasis:
		b = BASISFUNCTION_GAUSS;
		break;
	default:
		setErrorString("Detected invalid basis function setting");
		return false;
	}

	if (!si.writeInt32(b))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t n = (m_allowNegative)?1:0;

	if (!si.writeInt32(n))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t baseFlag = 0;
	
	if (m_pBaseLens)
		baseFlag = 1;
	
	if (!si.writeInt32(baseFlag))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (m_pBaseLens)
	{
		if (!m_pBaseLens->write(si))
		{
			setErrorString(m_pBaseLens->getErrorString());
			return false;
		}
	}

	int32_t searchType = SHEETSEARCHTYPE_NONE;

	switch(m_massSheetSearchType)
	{
	case Genome:
		searchType = SHEETSEARCHTYPE_GENOME;
		break;
	case None:
	default:
		searchType = SHEETSEARCHTYPE_NONE;
	}

	if (!si.writeInt32(searchType))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t evolutionFlag = 0;

	if (m_showEvolution)
		evolutionFlag |= 1;
	if (m_saveFirstGenome)
		evolutionFlag |= 2;

	if (!si.writeInt32(evolutionFlag))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	return true;
}

bool GridLensInversionGAFactoryParamsNew::read(serut::SerializationInterface &si)
{
	if (m_deleteImages)
	{
		while (!m_images.empty())
		{
			ImagesDataExtended *pImg = *(m_images.begin());

			delete pImg;

			m_images.pop_front();
		}
	}
	m_gridSquares.clear();
	m_images.clear();

	if (m_pBaseLens)
	{
		delete m_pBaseLens;
		m_pBaseLens = 0;
	}
	
	m_deleteImages = true;
	
	int32_t maxGen;

	if (!si.readInt32(&maxGen))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	m_maxGenerations = maxGen;
	
	if (!si.readDouble(&m_lensDistance))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.readDouble(&m_lensRedshift))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.readDouble(&m_initialMassScale))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t sizes[2];

	if (!si.readInt32s(sizes, 2))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int numImages = sizes[0];
	int numSquares = sizes[1];
	
	for (int i = 0 ; i < numImages ; i++)
	{
		ImagesDataExtended *pImg = new ImagesDataExtended();

		if (!pImg->read(si))
		{
			setErrorString(pImg->getErrorString());
			delete pImg;
			return false;
		}
		m_images.push_back(pImg);
	}

	for (int i = 0 ; i < numSquares ; i++)
	{
		double info[3];

		if (!si.readDoubles(info, 3))
		{
			setErrorString(si.getErrorString());
			return false;
		}
		m_gridSquares.push_back(GridSquare(Vector2D<double>(info[1], info[2]), info[0]));
	}
	
	int32_t b;

	if (!si.readInt32(&b))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	switch(b)
	{
	case BASISFUNCTION_PLUMMER:
		m_basisFunctionType = PlummerBasis;
		break;
	case BASISFUNCTION_SQUARE:
		m_basisFunctionType = SquareBasis;
		break;
	case BASISFUNCTION_GAUSS:
		m_basisFunctionType = GaussBasis;
		break;
	default:
		setErrorString("Read invalid basis function identifier");
		return false;
	}

	int32_t negativeFlag;

	if (!si.readInt32(&negativeFlag))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (negativeFlag == 0)
		m_allowNegative = false;
	else 
		m_allowNegative = true;

	int32_t baseLensFlag;

	if (!si.readInt32(&baseLensFlag))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (baseLensFlag != 0)
	{
		std::string errStr;

		if (!GravitationalLens::read(si, &m_pBaseLens, errStr))
		{
			setErrorString(errStr);
			return false;
		}
	}

	int32_t searchType;

	if (!si.readInt32(&searchType))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	switch(searchType)
	{
	case SHEETSEARCHTYPE_NONE:
		m_massSheetSearchType = None;
		break;
	case SHEETSEARCHTYPE_GENOME:
		m_massSheetSearchType = Genome;
		break;
	default:
		setErrorString("Invalid mass-sheet search type");
		return false;
	}

	int32_t evolutionFlag;
	
	if (!si.readInt32(&evolutionFlag))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if ((evolutionFlag&1) == 0)
		m_showEvolution = false;
	else 
		m_showEvolution = true;

	if ((evolutionFlag&2) == 0)
		m_saveFirstGenome = false;
	else
		m_saveFirstGenome = true;

	return true;
}
	
GridLensInversionGAFactoryParamsNew *GridLensInversionGAFactoryParamsNew::createCopy(bool copyImages) const
{
	GridLensInversionGAFactoryParamsNew *pParams = new GridLensInversionGAFactoryParamsNew();

	pParams->copyFieldsFrom(this, copyImages);

	return pParams;
}

void GridLensInversionGAFactoryParamsNew::copyFieldsFrom(const GridLensInversionGAFactoryParamsNew *pSrc, bool copyImages)
{
	if (m_deleteImages)
	{
		while (!m_images.empty())
		{
			ImagesDataExtended *pImg = *(m_images.begin());

			delete pImg;

			m_images.pop_front();
		}
	}
	m_gridSquares.clear();
	m_images.clear();

	if (m_pBaseLens)
	{
		delete m_pBaseLens;
		m_pBaseLens = 0;
	}

	m_maxGenerations = pSrc->m_maxGenerations;
	m_lensDistance = pSrc->m_lensDistance;
	m_lensRedshift = pSrc->m_lensRedshift;
	m_initialMassScale = pSrc->m_initialMassScale;
	m_gridSquares = pSrc->m_gridSquares;
	m_allowNegative = pSrc->m_allowNegative;
	m_basisFunctionType = pSrc->m_basisFunctionType;

	if (pSrc->m_pBaseLens)
		m_pBaseLens = pSrc->m_pBaseLens->createCopy(); // TODO: same as above

	m_massSheetSearchType = pSrc->m_massSheetSearchType;

	if (copyImages)
	{
		m_deleteImages = true;

		std::list<ImagesDataExtended *>::const_iterator it;

		for (it = pSrc->m_images.begin() ; it != pSrc->m_images.end() ; it++)
		{
			ImagesDataExtended *pImg = *it;

			m_images.push_back(new ImagesDataExtended(*pImg));
		}
	}
	else
	{
		m_deleteImages = false;
		m_images = pSrc->m_images;
	}

	m_showEvolution = pSrc->m_showEvolution;
	m_saveFirstGenome = pSrc->m_saveFirstGenome;
}

} // end namespace

