/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "treelensinversiongafactoryparams.h"
#include "imagesdataextended.h"

#include "debugnew.h"

namespace grale
{

TreeLensInversionGAFactoryParams::TreeLensInversionGAFactoryParams() : LensInversionGAFactoryParams(LensInversionGAFactoryParams::TreeInversion)
{
	m_deleteImages = false;
	m_Dd = 0;
	m_maxGenerations = 0;
	m_massScale = 0;
	m_basisFunctionType = PlummerBasis;
	m_areaSize = 0;
	m_maxBasisFunctions = 0;
	m_maxDepth = 0;
	m_useWeights = false;
	m_firstSubdiv = 2;
}

TreeLensInversionGAFactoryParams::TreeLensInversionGAFactoryParams(int maxGenerations, const std::list<ImagesDataExtended *> &images, 
		                                                   Vector2D<double> center, double areaSize,
								   int maxDepth, int maxBasisFunctions, 
								   double D_d, double massScale, bool copyImages, 
								   double widthScaleFactor,
								   bool useWeights,
								   int firstSubdiv,
								   BasisFunctionType b) : LensInversionGAFactoryParams(LensInversionGAFactoryParams::TreeInversion)
{
	if (copyImages)
		m_deleteImages = true;
	else
		m_deleteImages = false;

	m_maxGenerations = maxGenerations;
	m_center = center;
	m_areaSize = areaSize;
	m_maxDepth = maxDepth;
	m_maxBasisFunctions = maxBasisFunctions;
	m_Dd = D_d;
	m_massScale = massScale;
	m_basisFunctionType = b;
	m_widthScaleFactor = widthScaleFactor;
	m_useWeights = useWeights;
	m_firstSubdiv = firstSubdiv;

	std::list<ImagesDataExtended *>::const_iterator it;

	for (it = images.begin() ; it != images.end() ; it++)
	{
		ImagesDataExtended *pImg = (*it);
		ImagesDataExtended *pImg2;

		if (copyImages)
			pImg2 = new ImagesDataExtended(*pImg);
		else
			pImg2 = pImg;
			
		m_images.push_back(pImg2);
	}
}

TreeLensInversionGAFactoryParams::~TreeLensInversionGAFactoryParams()
{
	if (m_deleteImages)
	{
		while (!m_images.empty())
		{
			delete (*(m_images.begin()));
			m_images.pop_front();
		}
	}
}

#define BASISFUNCTION_PLUMMER 1
#define BASISFUNCTION_SQUARE  2
#define BASISFUNCTION_GAUSS   3

bool TreeLensInversionGAFactoryParams::write(serut::SerializationInterface &si) const
{
	if (!si.writeInt32(m_maxGenerations))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDouble(m_Dd))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDouble(m_massScale))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDoubles(m_center.getComponents(), 2))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDouble(m_areaSize))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDouble(m_widthScaleFactor))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeInt32(m_maxDepth))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeInt32(m_maxBasisFunctions))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t useWeights = (m_useWeights)?1:0;

	if (!si.writeInt32(useWeights))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (!si.writeInt32(m_firstSubdiv))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (!si.writeInt32(m_images.size()))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	std::list<ImagesDataExtended *>::const_iterator it;

	for (it = m_images.begin() ; it != m_images.end() ; it++)
	{
		if (!(*it)->write(si))
		{
			setErrorString((*it)->getErrorString());
			return false;
		}
	}

	int32_t b = 0;

	switch(m_basisFunctionType)
	{
	case PlummerBasis:
		b = BASISFUNCTION_PLUMMER;
		break;
	default:
		setErrorString("Detected invalid basis function setting");
		return false;
	}

	if (!si.writeInt32(b))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	return true;
}

bool TreeLensInversionGAFactoryParams::read(serut::SerializationInterface &si)
{
	if (m_deleteImages)
	{
		while (!m_images.empty())
		{
			delete (*(m_images.begin()));
			m_images.pop_front();
		}
	}
	
	m_images.clear();
	m_deleteImages = true;
	
	int32_t dummy;

	if (!si.readInt32(&dummy))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	m_maxGenerations = (int)dummy;
	
	if (!si.readDouble(&m_Dd))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.readDouble(&m_massScale))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.readDoubles(m_center.getComponents(), 2))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.readDouble(&m_areaSize))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.readDouble(&m_widthScaleFactor))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	
	if (!si.readInt32(&dummy))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	m_maxDepth = (int)dummy;

	if (!si.readInt32(&dummy))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	m_maxBasisFunctions = (int)dummy;

	if (!si.readInt32(&dummy))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (dummy == 0)
		m_useWeights = false;
	else
		m_useWeights = true;

	if (!si.readInt32(&dummy))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	m_firstSubdiv = (int)dummy;

	if (!si.readInt32(&dummy))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	
	int32_t numImages = dummy;
	
	for (int32_t i = 0 ; i < numImages ; i++)
	{
		ImagesDataExtended *pImg = new ImagesDataExtended();
		if (!pImg->read(si))
		{
			setErrorString(pImg->getErrorString());
			delete pImg;
			return false;
		}
		m_images.push_back(pImg);
	}

	int32_t b;

	if (!si.readInt32(&b))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	switch(b)
	{
	case BASISFUNCTION_PLUMMER:
		m_basisFunctionType = PlummerBasis;
		break;
	default:
		setErrorString("Read invalid basis function identifier");
		return false;
	}

	return true;
}
	
} // end namespace

