/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "wcscoordinates.h"
#ifndef WIN32
#include "thirdparty/wcssubs/wcs.h"
#endif // WIN32

#include "debugnew.h"

namespace grale
{

#ifndef WIN32

WCSCoordinates::WCSCoordinates()
{ 
	m_pCoords = 0; 
}

WCSCoordinates::~WCSCoordinates()
{ 
	if (m_pCoords) 
	{
		struct WorldCoor *pWC = (struct WorldCoor *)m_pCoords;
		wcsfree(pWC); 
	}
}

bool WCSCoordinates::init(const std::string &fitsHeader)
{
	if (m_pCoords)
	{
		setErrorString("WCS coordinate system is already initialized");
		return false;
	}

	m_pCoords = (void *)wcsninit(fitsHeader.c_str(), fitsHeader.size());
	if (!m_pCoords)
	{
		setErrorString("Unable to initialize the coordinate system");
		wcserr();
		return false;
	}

	m_fitsHeader = fitsHeader;
	return true;
}

CoordinateInterface *WCSCoordinates::createCopy() const
{
	WCSCoordinates *pNewInst = new WCSCoordinates();

	if (!pNewInst->init(m_fitsHeader))
	{
		setErrorString(pNewInst->getErrorString());
		delete pNewInst;
		return 0;
	}

	return pNewInst;
}

// TODO: I think the corner is (1,1) and not (0,0), but need to check this
Vector2D<double> WCSCoordinates::operator()(IntVector2D pos) const
{
	double xpos = 0;
	double ypos = 0;
	
	pix2wcs((struct WorldCoor *)m_pCoords, pos.getX()+1, pos.getY()+1, &xpos, &ypos);
	return Vector2D<double>((double)xpos,(double)ypos)*(double)ANGLE_DEGREE;
}
#else

WCSCoordinates::WCSCoordinates()
{ 
	m_pCoords = 0; 
}

WCSCoordinates::~WCSCoordinates()
{ 
}

bool WCSCoordinates::init(const std::string &fitsHeader)
{
	setErrorString("WCS coordinates are not available on the Win32 platform");
	return false;
}

CoordinateInterface *WCSCoordinates::createCopy() const
{
	setErrorString("WCS coordinates are not available on the Win32 platform");
	return 0;
}

// TODO: I think the corner is (1,1) and not (0,0), but need to check this
Vector2D<double> WCSCoordinates::operator()(IntVector2D pos) const
{
	return Vector2D<double>(0, 0);
}
#endif // WIN32

} // end namespace


