/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "imageplanegnuplotwriter.h"
#include "imageplane.h"
#include <stdio.h>

#include "debugnew.h"

namespace grale
{

bool ImagePlaneGnuPlotWriter::write(ImagePlane *p, int subSamples)
{
	FILE *f = fopen(filename.c_str(),"w");
	
	if (f == 0)
	{
		setErrorString("Can't open file");
		return false;
	}

	double pointsize = 70.0/(double)(p->getNumXPixels());
	if (pointsize < 0.5)
		pointsize = 0.5;
	
	if (col)
		fprintf(f,"set terminal postscript eps color solid enhanced %d\n",fontsize);
	else
		fprintf(f,"set terminal postscript eps monochrome enhanced %d\n",fontsize);
	fprintf(f,"set size 1.0,1.3\n");
	fprintf(f,"set style data points\n");
	fprintf(f,"set pointsize %g\n",pointsize);
	fprintf(f,"set xlabel '%s'\n",Xlabel.c_str());
	fprintf(f,"set ylabel '%s'\n",Ylabel.c_str());
	if (grid)
		fprintf(f,"set grid\n");
	
	int pointtype = 31;
	int count = 0;
	bool first = true;
	
	if (im) count++;
	if (srcs) count++;
	if (crit) count++;
	if (caust) count++;

	Vector2D<double> topright = p->getTopRight()/scale;
	Vector2D<double> bottomleft = p->getBottomLeft()/scale;
	
	if (m_flipX)
		fprintf(f,"plot [%Lg:%Lg][%Lg:%Lg] ",(long double)topright.getX(),(long double)bottomleft.getX(),(long double)bottomleft.getY(),(long double)topright.getY());
	else
		fprintf(f,"plot [%Lg:%Lg][%Lg:%Lg] ",(long double)bottomleft.getX(),(long double)topright.getX(),(long double)bottomleft.getY(),(long double)topright.getY());
	
	if (count == 0)
	{
		fprintf(f,"%Lg notitle\n",(long double)(bottomleft.getY()*2.0));
		fclose(f);
		return true;
	}

	if (caust)
	{
		first = false;
		if (col)
			fprintf(f,"'-' t 'Caustics' with linespoints lw 3 lt 3 pt 1 ps 0.05");
		else
			fprintf(f,"'-' t 'Caustics' with linespoints lw 3 lt 1 pt 1 ps 0.05");
	}
	if (crit)
	{
		if (!first)
			fprintf(f,",");
		else
			first = false;
		
		if (col)
			fprintf(f,"'-' t 'Critical lines' with linespoints lw 3 lt 1 pt 1 ps 0.05");
		else
			fprintf(f,"'-' t 'Critical lines' with lines lw 3 lt 3");
	}
	if (srcs)
	{
		if (!first)
			fprintf(f,",");
		else
			first = false;
		
		if (col)
			fprintf(f,"'-' t 'Sources' pt %d lt 9",pointtype);
		else
			fprintf(f,"'-' t 'Sources' pt %d lt 1",pointtype);
	}
	if (im)
	{
		if (!first)
			fprintf(f,",");
		else
			first = false;
		
		if (col)
			fprintf(f,"'-' t 'Images' pt %d lt 0",pointtype);
		else
			fprintf(f,"'-' t 'Images' pt %d lt 1",pointtype);
	}
	fprintf(f,"\n");

	if (caust)
	{
		const std::list<std::list<Vector2D<double> > *> &caust2 = p->getCausticSegments();
		const std::list<Vector2D<double> > *line;
		std::list<std::list<Vector2D<double> > *>::const_iterator it1;
		std::list<Vector2D<double> >::const_iterator it2;

		for (it1 = caust2.begin() ; it1 != caust2.end() ; it1++)
		{
			if (it1 != caust2.begin())
				fprintf(f,"\n"); // start new line segment
			line = *it1;

			for (it2 = line->begin() ; it2 != line->end() ; it2++)
			{
				Vector2D<double> pos = *it2;
				pos /= scale;
				fprintf(f,"%30.30Lf\t %30.30Lf\n",(long double)pos.getX(),(long double)pos.getY());
			}
		}
		if (caust2.empty())
		{
			fprintf(f,"%30.30Lf\t %30.30Lf\n",(long double)(topright.getX()*2.0L),(long double)(topright.getY()*2.0L));
		}	
		fprintf(f,"e\n");
	}
	if (crit)
	{
		const std::list<std::list<Vector2D<double> > *> &crlines2 = p->getCriticalLineSegments();
		const std::list<Vector2D<double> > *line;
		std::list<std::list<Vector2D<double> > *>::const_iterator it1;
		std::list<Vector2D<double> >::const_iterator it2;

		for (it1 = crlines2.begin() ; it1 != crlines2.end() ; it1++)
		{
			if (it1 != crlines2.begin())
				fprintf(f,"\n"); // start new line
			line = *it1;

			for (it2 = line->begin() ; it2 != line->end() ; it2++)
			{
				Vector2D<double> pos = *it2;
				pos /= scale;
				fprintf(f,"%30.30Lf\t %30.30Lf\n",(long double)pos.getX(),(long double)pos.getY());
			}
		}
		if (crlines2.empty())
		{
			fprintf(f,"%30.30Lf\t %30.30Lf\n",(long double)topright.getX()*2.0L,(long double)topright.getY()*2.0L);
		}
		fprintf(f,"e\n");
	}
	if (srcs)
	{
		int x,y,numx,numy;
		bool hit = false;

		numx = p->getNumXPixels();
		numy = p->getNumYPixels();

		for (y = 0 ; y < numy ; y++)
		{
			for (x = 0 ; x < numx ; x++)
			{
				double value = p->getSourceIntensityAccurate(x,y,subSamples); 

				if (value > 0)
				{
					hit = true;
					Vector2D<double> pos = p->getPixelCoordinate(x,y);
					pos /= scale;
					fprintf(f,"%30.30g\t %30.30g\t %30.30g\n", pos.getX(), pos.getY(), value);
				}
			}
		}
		if (!hit)
		{
			fprintf(f,"%30.30g\t %30.30g\n", topright.getX()*2.0, topright.getY()*2.0);
		}
		fprintf(f,"e\n");
	}
	if (im)
	{
		int x,y,numx,numy;
		bool hit = false;

		numx = p->getNumXPixels();
		numy = p->getNumYPixels();

		for (y = 0 ; y < numy ; y++)
		{
			for (x = 0 ; x < numx ; x++)
			{
				double value = p->getImageIntensityAccurate(x,y,subSamples);

				if (value > 0)
				{
					hit = true;
					Vector2D<double> pos = p->getPixelCoordinate(x,y);
					pos /= scale;
					fprintf(f,"%30.30g\t %30.30g\t %30.30g\n", pos.getX(), pos.getY(), value);
				}
			}
		}
		if (!hit)
		{
			fprintf(f,"%30.30g\t %30.30g\n", topright.getX()*2.0, topright.getY()*2.0);
		}
		fprintf(f,"e\n");
	}

	fclose(f);
	return true;
}

} // end nalespace

