/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"

#ifdef GRALECONFIG_SUPPORT_CFITSIO

#include "real2dfunctionfitsplotter.h"
#include <fitsio.h>
#ifndef WIN32
#include <unistd.h>
#endif // WIN32
#include <sys/types.h>
#include <stdint.h>
#include <sys/stat.h>

#include "debugnew.h"

namespace grale
{

Real2DFunctionFITSPlotter::Real2DFunctionFITSPlotter(const std::string &fname,double xs,double ys,double zs)
{
	filename = fname;
	xscale = xs;
	yscale = ys;
	zscale = zs;
}

Real2DFunctionFITSPlotter::~Real2DFunctionFITSPlotter()
{
}

bool Real2DFunctionFITSPlotter::plot(Real2DFunction &f,Vector2D<double> bottomleft,Vector2D<double> topright,int numxsteps,int numysteps)
{
	if (numxsteps < 2 || numysteps < 2)
	{
		setErrorString("At least 2 steps in X and Y direction are required");
		return false;
	}
	
	struct stat fileinfo;
	fitsfile *fptr;
	int status = 0;

	if (stat(filename.c_str(),&fileinfo) == 0) // file exists, remove it
	{
		if (unlink(filename.c_str()) != 0)
		{
			setErrorString("Can't remove existing file");
			return false;
		}
	}
			
	if (fits_create_file(&fptr,filename.c_str(),&status))
	{
		//fits_report_error(stderr,status);
		setErrorString("Can't create new FITS file");
		return false;
	}
	
	int bitpix;
	bitpix = DOUBLE_IMG;

	long naxis = 2;
	long naxes[2] = {numxsteps,numysteps};
	
	if (fits_create_img(fptr,bitpix,naxis,naxes,&status))
	{
		//fits_report_error(stderr,status);
		fits_close_file(fptr,&status);
		setErrorString("Can't create FITS image entry");
		return false;
	}

	int datalen = numxsteps*numysteps;
	double *data = new double[datalen];
	int index = 0;
	int x,y;
	double startx = bottomleft.getX();
	double starty = bottomleft.getY();
	double xstep = (topright.getX()-bottomleft.getX())/((double)(numxsteps-1));
	double ystep = (topright.getY()-bottomleft.getY())/((double)(numysteps-1));
	
	for (y = 0 ; y < numysteps ; y++)
	{
		for (x = 0 ; x < numxsteps ; x++)
		{
			double z = f(Vector2D<double>(startx+xstep*(double)x,starty+ystep*(double)y));
			data[index] = (double)(z/zscale);
			index++;
		}
	}

	if (fits_write_img(fptr,TDOUBLE,1,numxsteps*numysteps,data,&status))
	{
		delete [] data;
		//fits_report_error(stderr,status);
		fits_close_file(fptr,&status);
		setErrorString("Can't write FITS image data");
		return false;
	}
	
	delete [] data;
	
	double stepx = (double)(xstep/xscale);
	double stepy = (double)(ystep/yscale);
	double xstart = (double)(startx/xscale);
	double ystart = (double)(starty/yscale);
	
	if (fits_update_key(fptr, TDOUBLE, "CRVAL1", &xstart," ", &status))
	{
		//fits_report_error(stderr,status);
		fits_close_file(fptr,&status);
		setErrorString("Can't write FITS CRVAL1 key");
		return false;
	}
	if (fits_update_key(fptr, TDOUBLE, "CRVAL2", &ystart," ", &status))
	{
		//fits_report_error(stderr,status);
		fits_close_file(fptr,&status);
		setErrorString("Can't write FITS CRVAL2 key");
		return false;
	}	
	if (fits_update_key(fptr, TDOUBLE, "CDELT1", &stepx," ", &status))
	{
		//fits_report_error(stderr,status);
		fits_close_file(fptr,&status);
		setErrorString("Can't write FITS CDELT1 key");
		return false;
	}
	if (fits_update_key(fptr, TDOUBLE, "CDELT2", &stepy," ", &status))
	{
		//fits_report_error(stderr,status);
		fits_close_file(fptr,&status);
		setErrorString("Can't write FITS CDELT2 key");
		return false;
	}
	
	if (fits_close_file(fptr,&status))
	{
		//fits_report_error(stderr,status);
		setErrorString("Can't close FITS file");
		return false;
	}

	return true;
}

} // end namespace

#endif // GRALECONFIG_SUPPORT_CFITSIO

